from .util import SaveLoad


class Scanner(SaveLoad):
    """Base scanner class.

    Attributes
    ----------
    classes : `dict`
        Scanner class group.
    START
        Sentence start token.
    END
        Sentence end token.

    Examples
    --------
    >>> scan = Scanner(lambda data: data.split())
    >>> scan('a b c')
    ['a', 'b', 'c']
    """

    classes = {}

    END = None
    START = None

    def __init__(self, scan=None):
        """Base scanner constructor.

        Parameters
        ----------
        scan : `function`, optional
        """
        if scan is not None:
            self.do_scan = scan

    def __call__(self, data, part=False):
        """Scan data.

        Parameters
        ----------
        data
            Data to scan.
        part : `bool`, optional
            `True` if data is partial (default: `False`).

        Returns
        -------
        `object`
            self.scan(data)
        """
        if self.do_scan is None:
            return data
        return self.do_scan(data)

    def reset(self):
        """Reset scanner state.
        """
        pass



import sys
import math
from abc import abstractmethod
from itertools import chain, islice

from ..util import SaveLoad
from ..text.util import get_words


class Rank(SaveLoad):
    """Base text rank class.

    Attributes
    ----------
    size : `int`
    remove : `float`
    debug : `bool`
        If True, enable debug output.
    """
    classes = {}

    def __init__(self, size=10, remove=0.5):
        if size <= 0:
            raise ValueError('rank size <= 0')
        self.size = size
        self.remove = remove
        self.debug = False

    def __eq__(self, rank):
        return (super().__eq__(self, rank)
                and self.size == rank.size
                and abs(self.remove - rank.remove) < 1e-5)

    def save(self):
        ret = super().save()
        ret['size'] = self.size
        ret['remove'] = self.remove
        return ret

    @abstractmethod
    def rank(self, string):
        """Rank a string.

        Parameters
        ----------
        string : `str`

        Returns
        -------
        `float`
        """
        pass

    def __call__(self, strings):
        """Filter strings by rank.

        Parameters
        ----------
        strings : `iterable` of `str`
            Strings to filter.

        Returns
        -------
        `list` of `str`
            Filtered list.
        """
        strings = sorted(
            ((string, self.rank(string)) for string in strings),
            key=lambda x: -x[1]
        )
        end = max(1, len(strings) - int(self.remove * len(strings)))
        res = [string for string, rank in islice(strings, 0, end)]
        if self.debug:
            print(res, file=sys.stderr)
        return res


class Const(Rank):
    """Constant text rank."""
    def __init__(self, **_):
        super().__init__(1, 0.0)

    def rank(self, string):
        return 1


class Test(Rank):
    def __init__(self, size, remove):
        super().__init__(size=10, remove=0.5)
        self.header = False
        self.opt_words = 8
        self.opt_long_words = 4
        self.opt_long_word_ratio = 0.6
        self.long_word_length = 4

    def features(self, string):
        words = get_words(string)
        nwords = len(words)
        nlongwords = sum(
            1 for word in words
            if len(word) >= self.long_word_length
        )
        return [
            1 - abs(1 - nwords / self.opt_words),
            1 - abs(1 - nlongwords / self.opt_long_words) ** 2,
            #len(set(words)) / nwords,
            1 - abs(1 - nlongwords / nwords / self.opt_long_word_ratio)
        ]

    def log(self, res, features, string):
        if not self.header:
            self.header = True
            fmt = ' Rank     ' + '  %02d      ' * len(features)
            print(fmt % tuple(range(len(features))), file=sys.stderr)
        fmt = '%.04f    ' * (len(features) + 1) + '%s'
        print(fmt % tuple(chain((res,), features, (string,))), file=sys.stderr)

    def rank(self, string):
        features = self.features(string)
        for i, x in enumerate(features):
            features[i] = min(1, max(0, x))
        ret = sum(features) / len(features)
        if self.debug:
            self.log(ret, features, string)
        return ret

    def __call__(self, strings):
        self.header = False
        return super().__call__(strings)



import pytest
from PIL import Image

from markovchain import Scanner
from markovchain.image import ImageScanner, HLines, VLines


@pytest.fixture(scope='module')
def image_test():
    data = b'\x00\x02\x03\x01'
    image = Image.frombytes('P', (2, 2), data, 'raw', 'P')
    return image


def test_image_scanner_init():
    scan = ImageScanner()
    assert scan.levels is not None
    assert scan.min_size is not None
    assert isinstance(scan.level_scale, list)
    assert isinstance(scan.traversal, list)
    assert len(scan.level_scale) == scan.levels - 1
    assert len(scan.traversal) == scan.levels

def test_image_scanner_properties():
    scan = ImageScanner()

    with pytest.raises(ValueError):
        scan.levels = 0
    scan.levels = 3
    assert len(scan.traversal) == scan.levels

    with pytest.raises(ValueError):
        scan.level_scale = []
    with pytest.raises(ValueError):
        scan.level_scale = [1, -1]
    scan.level_scale = range(2, 10)
    assert scan.level_scale == [2, 3]
    assert scan.min_size == 6
    scan.level_scale = 2
    assert scan.level_scale == [2, 2]
    assert scan.min_size == 4
    scan.levels = 2
    assert len(scan.traversal) == scan.levels
    assert scan.level_scale == [2]
    assert scan.min_size == 2
    scan.levels = 1
    assert len(scan.traversal) == scan.levels
    assert scan.level_scale == []
    assert scan.min_size == 1

    traversal = [HLines(), VLines()]
    scan.traversal = traversal
    scan.levels = 2
    assert scan.traversal is traversal
    assert scan.level_scale == [2]

@pytest.mark.parametrize('test', [
    (1, 1), (4, 1), (1, 4)
])
def test_image_scanner_input(test, image_test):
    scan = ImageScanner(resize=test)
    img = scan.input(image_test)
    assert img.size <= test
    assert img.mode == 'P'
    img = [scan.level(img, level) for level in range(scan.levels)]
    assert len(img) == 1
    img = img[0]
    assert img.mode == 'P'
    assert img.size == (1, 1)
    assert list(img.getdata()) in [[0], [1]]

def test_image_scanner_input_error(image_test):
    scan = ImageScanner(levels=3, level_scale=2)
    with pytest.raises(ValueError):
        scan.input(image_test)

def test_image_scanner_input_levels(image_test):
    scan = ImageScanner(resize=(4, 4), levels=3, level_scale=2)

    img = scan.input(image_test)
    img = [scan.level(img, level) for level in range(scan.levels)]

    assert len(img) == 3

    assert img[0].mode == 'P'
    assert img[0].size == (1, 1)
    assert list(img[0].getdata()) in [[0], [1]]

    assert img[1].mode == 'P'
    assert img[1].size == (2, 2)
    assert list(img[1].getdata()) == [0, 2, 3, 1]

    assert img[2].mode == 'P'
    assert img[2].size == (4, 4)
    assert list(img[2].getdata()) == [
        0, 0, 2, 2,
        0, 0, 2, 2,
        3, 3, 1, 1,
        3, 3, 1, 1
    ]

def test_image_scanner_input_level_scale():
    img = Image.new(mode='RGB', size=(48, 48))
    scan = ImageScanner(levels=4, level_scale=[2, 3, 4])
    img = scan.input(img)
    size = [scan.level(img, level).size for level in range(scan.levels)]
    assert size == [(2, 2), (4, 4), (12, 12), (48, 48)]

def test_image_scanner_scan(image_test):
    scan = ImageScanner(traversal=HLines())
    assert [list(level) for level in scan(image_test)] == [
        ['\x00', '\x02', '\x03', '\x01', scan.END]
    ]

def test_image_scanner_scan_levels(image_test):
    scan = ImageScanner(levels=2, level_scale=2,
                        traversal=[HLines(), VLines()])
    assert [list(level) for level in scan(image_test)] in [
        [
            ['\x00', scan.END],
            [(scan.START, '\x00'),
             '\x00', '\x03', '\x02', '\x01',
             scan.END]
        ],
        [
            ['\x01', scan.END],
            [(scan.START, '\x01'),
             '\x00', '\x03', '\x02', '\x01',
             scan.END]
        ]
    ]

@pytest.mark.parametrize('test', [
    (),
    ((4, 4), 2, 2, Image.NEAREST, [HLines(), VLines()])
])
def test_image_scanner_save_load(test):
    scanner = ImageScanner(*test)
    saved = scanner.save()
    loaded = Scanner.load(saved)
    assert isinstance(loaded, ImageScanner)
    assert scanner == loaded



from unittest.mock import Mock, call
import pytest

from markovchain.storage.base import Storage


class StorageTest(Storage):
    def replace_state_separator(self, old_separator, new_separator):
        pass
    def get_dataset(self, key, create=False):
        pass
    def add_links(self, links, dataset_prefix=''):
        pass
    def get_state(self, state, size):
        pass
    def get_states(self, dataset, string):
        pass
    def get_links(self, dataset, state, backward=False):
        pass
    def follow_link(self, link, state, backward=False):
        pass
    def do_save(self, fp=None):
        pass
    @classmethod
    def load(cls, fp):
        pass

def test_storage_base_abstract():
    with pytest.raises(TypeError):
        Storage()

def test_storage_base_properties():
    storage = StorageTest()
    storage.replace_state_separator = Mock()
    assert storage.state_separator == ' '
    storage.state_separator = ':'
    storage.replace_state_separator.assert_called_once_with(' ', ':')
    assert storage.state_separator == ':'

@pytest.mark.parametrize('test,test2,res', [
    ((), (), True),
    (({'state_separator': '-'},), ({'state_separator': '+'},), False)
])
def test_storage_base_eq(test, test2, res):
    assert (StorageTest(*test) == StorageTest(*test2)) == res

@pytest.mark.parametrize('settings,test,res', [
    ({}, 'a b c', ['a', 'b', 'c']),
    ({'storage': {'state_separator': '-'}}, 'a b-c d', ['a b', 'c d']),
    ({'storage': {'state_separator': ''}}, 'a b', ['a', ' ', 'b']),
])
def test_storage_base_split_state(settings, test, res):
    assert StorageTest(settings).split_state(test) == res

@pytest.mark.parametrize('settings,test,res', [
    ({}, map(str, range(3)), '0 1 2'),
    ({'storage': {'state_separator': '-'}}, ['a', 'b'], 'a-b'),
])
def test_storage_base_join_state(settings, test, res):
    assert StorageTest(settings).join_state(test) == res

@pytest.mark.parametrize('args,random,res', [
    (('data', 'x', True), 0, 'y'),
    (('data', 'x', True), 1, 'z'),
    (('data', 'z', False), 0, 'x'),
    (('data', 'y', True), [], None)
])
def test_storage_base_random_link(mocker, args, random, res):
    randint = mocker.patch(
        'markovchain.storage.base.randint',
        return_value=random
    )
    links = {
        'x': [(1, 'y'), (1, 'z')],
        'y': [],
        'z': [(1, 'x')]
    }
    storage = StorageTest()
    storage.get_links = Mock(
        wraps=lambda dataset, node, backward: links.get(node, [])
    )
    storage.follow_link = Mock(return_value=0)

    test = storage.random_link(*args)
    s = sum(count for count, link in links.get(args[1], []))
    storage.get_links.assert_called_with(*args)
    if s > 0:
        assert test == (res, 0)
        randint.assert_called_with(0, s - 1)
        storage.follow_link.assert_called_with(
            links[args[1]][random],
            args[1],
            args[2]
        )
    else:
        assert test == (None, None)
        assert randint.call_count == 0
        assert storage.follow_link.call_count == 0

@pytest.mark.parametrize('state,backward', [
    ('x', True),
    (['x', 'y'], False)
])
def test_storage_base_generate(state, backward):
    storage = StorageTest()
    storage.random_link = Mock(side_effect=[
        ('link', 'state'), ('link2', 'state2'), (None, None)
    ])
    storage.get_dataset = Mock(return_value=2)
    storage.get_state = Mock(return_value=1)
    storage.split_state = Mock(return_value=3)

    res = list(storage.generate(state, 4, 'data', backward))
    assert res == ['link', 'link2']
    storage.get_dataset.assert_called_with('data')
    if isinstance(state, str):
        storage.split_state.assert_called_with(state)
        storage.get_state.assert_called_with(3, 4)
    else:
        assert storage.split_state.call_count == 0
        storage.get_state.assert_called_with(state, 4)
    assert storage.random_link.call_count == 3
    storage.random_link.assert_has_calls([
        call(2, 1, backward),
        call(2, 'state', backward),
        call(2, 'state2', backward)
    ])

def test_storage_base_save():
    storage = StorageTest()
    storage.do_save = Mock()
    storage.state_separator = '+'
    storage.save(0)
    assert storage.settings['storage']['state_separator'] == '+'
    storage.do_save.assert_called_once_with(0)



import sys
import json
from collections import deque
from itertools import chain, repeat, tee

from .base import Storage


class JsonStorage(Storage):
    """JSON storage.

    Attributes
    ----------
    nodes : `dict` of `dict` of ([`int`, `str`] or [`list` of `int`, `list` of `str`])
    backward : `None` or `dict` of `dict` of ([`int`, `str`] or [`list` of `int`, `list` of `str`])
    """
    def __init__(self, nodes=None, backward=None, settings=None):
        """JSON storage constructor.

        Parameters
        ----------
            nodes : `dict` of `dict` of ([`int`, `str`] or [`list` of `int`, `list` of `str`]), optional
            backward : `bool` or `dict` of `dict` of ([`int`, `str`] or [`list` of `int`, `list` of `str`]), optional
        """
        if nodes is None:
            nodes = {}

        if backward is None:
            if settings is not None:
                if settings.get('storage', {}).get('backward', False):
                    backward = {}
                else:
                    backward = None
        elif isinstance(backward, bool):
            if backward:
                backward = {}
            else:
                backward = None

        super().__init__(settings)
        self.nodes = nodes
        self.backward = backward

    def __eq__(self, storage):
        return (self.nodes == storage.nodes
                and self.backward == storage.backward
                and super().__eq__(storage))

    @staticmethod
    def do_replace_state_separator(data, old, new):
        """Replace state separator.

        Parameters
        ----------
        data : `dict` of `dict` of ([`int`, `str`] or [`list` of `int`, `list` of `str`])
            Data.
        old : `str`
            Old separator.
        new : `str`
            New separator.
        """
        for key, dataset in data.items():
            data[key] = dict(
                (k.replace(old, new), v)
                for k, v in dataset.items()
            )

    @staticmethod
    def do_get_dataset(data, key, create=False):
        """Get a dataset.

        Parameters
        ----------
        data : `None` or `dict` of `dict` of ([`int`, `str`] or [`list` of `int`, `list` of `str`])
            Data.
        key : `str`
            Dataset key.
        create : `bool`, optional
            Create a dataset if it does not exist.

        Returns
        -------
        `None` or `dict` of ([`int`, `str`] or [`list` of `int`, `list` of `str`])
        """
        if data is None:
            return None
        try:
            return data[key]
        except KeyError:
            if create:
                dataset = {}
                data[key] = dataset
                return dataset
            else:
                raise

    @staticmethod
    def add_link(dataset, source, target, count=1):
        """Add a link.

        Parameters
        ----------
        dataset : `dict` of ([`int`, `str`] or [`list` of `int`, `list` of `str`])
            Dataset.
        source : `iterable` of `str`
            Link source.
        target : `str`
            Link target.
        count : `int`, optional
            Link count (default: 1).
        """
        try:
            node = dataset[source]
            values, links = node
            if isinstance(links, list):
                try:
                    idx = links.index(target)
                    values[idx] += count
                except ValueError:
                    links.append(target)
                    values.append(count)
            elif links == target:
                node[0] += count
            else:
                node[0] = [values, count]
                node[1] = [links, target]
        except KeyError:
            dataset[source] = [count, target]

    def replace_state_separator(self, old_separator, new_separator):
        self.do_replace_state_separator(
            self.nodes,
            old_separator,
            new_separator
        )
        if self.backward is not None:
            self.do_replace_state_separator(
                self.backward,
                old_separator,
                new_separator
            )

    def get_dataset(self, key, create=False):
        return (
            self.do_get_dataset(self.nodes, key, create),
            self.do_get_dataset(self.backward, key, create)
        )

    def add_links(self, links, dataset_prefix=''):
        for dataset, src, dst in links:
            forward, backward = self.get_dataset(dataset_prefix + dataset, True)
            if backward is not None and dst is not None:
                src, src2 = tee(src)
                dst2 = next(src2)
                src2 = self.join_state(chain(src2, (dst,)))
                self.add_link(backward, src2, dst2)
            src = self.join_state(src)
            self.add_link(forward, src, dst)

    def get_state(self, state, size):
        return deque(chain(repeat('', size), state), maxlen=size)

    def get_states(self, dataset, string):
        dataset = self.get_dataset(dataset)[0]
        string = string.lower()
        return [key for key in dataset.keys() if string in key.lower()]

    def get_links(self, dataset, state, backward=False):
        """
        Raises
        ------
        ValueError
            If backward == `True` and self.backward is `None`.
        """
        if backward and self.backward is None:
            raise ValueError('no backward nodes')
        try:
            node = dataset[int(backward)][self.join_state(state)]
            if not isinstance(node[0], list):
                return [(node[0], node[1])]
            return list(zip(*node))
        except KeyError:
            return []

    def follow_link(self, link, state, backward=False):
        value = link[1]
        if backward:
            state.appendleft(value)
        else:
            state.append(value)
        return state

    def do_save(self, fp=None):
        """Save to file.

        Parameters
        ----------
        fp : `file` or `str`, optional
            Output file (default: stdout).
        """
        if fp is None:
            fp = sys.stdout
        data = {
            'settings': self.settings,
            'nodes': self.nodes,
            'backward': self.backward
        }
        json.dump(data, fp, ensure_ascii=False)

    @classmethod
    def load(cls, fp):
        if isinstance(fp, str):
            with open(fp, 'rt') as fp2:
                data = json.load(fp2)
        else:
            data = json.load(fp)
        return cls(**data)



import re
import pytest

from markovchain.text.util import (
    CharCase, ispunct, capitalize, lstrip_ws_and_chars,
    re_flags, re_flags_str, re_sub, get_words, ReFlags
)


@pytest.mark.parametrize('test,res', [
    ('\'"?,+-.[]{}()<>', True),
    ('\'"?,+-x.[]{}()<>', False),
    ('', False)
])
def test_ispunct(test, res):
    assert ispunct(test) == res


@pytest.mark.parametrize('test,res', [
    ('worD WORD WoRd', 'Word word word'),
    ('x', 'X'),
    ('', '')
])
def test_capitalize(test, res):
    assert capitalize(test) == res


@pytest.mark.parametrize('test,res', [
    (('', ''), ''),
    (('     ', ''), ''),
    (('  x  ', 'xy'), ''),
    ((' \t.\n , .x. ', '.,?!'), 'x. ')
])
def test_lstrip_ws_and_chars(test, res):
    assert lstrip_ws_and_chars(*test) == res


@pytest.mark.parametrize('case,string,res', [
    (CharCase.PRESERVE, 'a B c', 'a B c'),
    (CharCase.TITLE, 'a B c', 'A b c'),
    (CharCase.UPPER, 'a B c', 'A B C'),
    (CharCase.LOWER, 'a B c', 'a b c')
])
def test_char_case_convert(case, string, res):
    case = CharCase(case)
    assert case.convert(string) == res


@pytest.mark.parametrize('test,res', [
    ('ab,cd', ['ab', 'cd']),
    ('ab cd..\n,ef', ['ab', 'cd', 'ef'])
])
def test_get_words(test, res):
    assert get_words(test) == res


@pytest.mark.parametrize('test,res', [
    (('uiM',), (re.U | re.I | re.M, 0)),
    (('UO',), (re.U, ReFlags.O)),
    (('O', None), ValueError),
    (('-',), ValueError)
])
def test_re_flags(test, res):
    if isinstance(res, type):
        with pytest.raises(res):
            re_flags(*test)
    else:
        assert re_flags(*test) == res


@pytest.mark.parametrize('test,res', [
    ((re.U | re.I | re.M, 0), 'UIM'),
    ((re.U, ReFlags.O), 'UO')
])
def test_re_flags_str(test, res):
    assert sorted(re_flags_str(*test)) == sorted(res)


@pytest.mark.parametrize('test,res', [
    (('x+', 'y', 'xxzxxx'), 'yzy'),
    (('x+', 'y', 'xxzxxx', 1), 'yzxxx'),
    (('xx', 'y', 'xxzXXX', 0, re.I), 'yzyX'),
    (('xx', 'x', 'xxzXXX', 0, re.I, ReFlags.O), 'xzx')
])
def test_re_sub(test, res):
    assert re_sub(*test) == res



from collections import Counter
from itertools import islice
import pytest

from markovchain.image.util import palette


@pytest.mark.parametrize('test', [
    (8, 8, 5),
    (0, 4, 4),
    (4, 0, 4),
    (4, 4, 0),
    (8, -4, 8),
    (-8, 4, -8)
])
def test_palette_errors(test):
    with pytest.raises(ValueError):
        palette(*test)

@pytest.mark.parametrize('test', [
    (8, 4, 8),
    (2, 2, 2),
    (16, 1, 1)
])
def test_palette(test):
    res = palette(*test)
    assert len(res) == 768
    res = list(zip(islice(res, 0, None, 3),
                   islice(res, 1, None, 3),
                   islice(res, 2, None, 3)))
    counter = Counter(res)
    size = test[0] * test[1] * test[2]
    assert len(counter.items()) == min(256, size + 1)
    assert counter[(0, 0, 0)] == 256 - size

@pytest.mark.parametrize('test', [
    (1, 1, 1),
    (1, 1, 2),
    (1, 1, 256)
])
def test_palette_grayscale(test):
    res = palette(*test)
    assert len(res) == 768
    res = list(zip(islice(res, 0, None, 3),
                   islice(res, 1, None, 3),
                   islice(res, 2, None, 3)))
    assert all(r == g and g == b for r, g, b in res)
    counter = Counter(res)
    size = test[0] * test[1] * test[2]
    assert len(counter.items()) == size
    assert counter[(0, 0, 0)] == 256 - size + 1
    if size > 1:
        assert counter[(255, 255, 255)] == 1



import random
from itertools import chain, islice, cycle

from .formatter import FormatterBase, Formatter
from .rank import Rank, Const
from .scanner import RegExpScanner
from .util import get_words, ReplyMode
from ..parser import Parser
from ..base import Markov
from ..util import load, state_size_dataset


class MarkovText(Markov):
    """Markov text generator class.
    """

    DEFAULT_SCANNER = RegExpScanner
    DEFAULT_PARSER = Parser
    DEFAULT_FORMATTER = Formatter
    DEFAULT_RANK = Const

    def __init__(self,
                 scanner=None,
                 parser=None,
                 storage=None,
                 formatter=None,
                 rank=None):
        super().__init__(scanner, parser, storage)
        self.rank = load(rank, Rank, self.DEFAULT_RANK)
        self.formatter = load(formatter, FormatterBase, self.DEFAULT_FORMATTER)

    def __eq__(self, markov):
        return (super().__eq__(markov)
                and self.rank == markov.rank
                and self.formatter == markov.formatter)

    def get_settings_json(self):
        data = super().get_settings_json()
        data['rank'] = self.rank.save()
        data['formatter'] = self.formatter.save()
        return data

    def data(self, data, part=False, dataset=''):
        """
        Parameters
        ----------
        data : `str`
            Text to parse.
        part : `bool`, optional
            True if data is partial (default: `False`).
        dataset : `str`, optional
            Dataset key prefix (default: '').
        """
        return super().data(data, part)

    def format(self, parts):
        """Format generated text.

        Parameters
        ----------
        parts : `iterable` of `str`
            Text parts.
        """
        text = self.storage.state_separator.join(parts)
        return self.formatter(text)

    def get_cont_state(self, string, backward=False):
        """Get initial states from input string.

        Parameters
        ----------
        string : `str` or `None`
        backward : `bool`

        Returns
        -------
        `tuple` of `str`
        """
        if string is None:
            return ()
        for _ in self.parser(self.scanner(string, True), True):
            if backward and len(self.parser.state[0]):
                break
        state = tuple(self.parser.state)
        self.scanner.reset()
        self.parser.reset()
        return state

    def get_reply_states(self, string, dataset):
        """Get initial states from input string.

        Parameters
        ----------
        string : `str`
            Input string.
        dataset : `str`
            Dataset key.

        Returns
        -------
        `list` of `list` of `str`
        """
        words = get_words(string)
        if not words:
            return []
        long_word = 4
        long_words = [word for word in words if len(word) >= long_word]
        short_words = [word for word in words if len(word) < long_word]
        for words in (long_words, short_words):
            ret = [
                states
                for states in (
                    self.storage.get_states(dataset, word)
                    for word in words
                )
                if states
            ]
            if ret:
                return ret
        return []

    def generate_cont(self, max_length, state_size,
                      reply_to, backward, dataset):
        """Generate texts from start/end.

        Parameters
        ----------
        max_length : `int` or `None`
            Maximum sentence length.
        state_size : `int`
            State size.
        reply_to : `str` or `None`
            Input string.
        backward : `bool`
            `True` to generate text start.
        dataset: `str`
            Dataset key prefix.

        Returns
        -------
        `generator` of `str`
            Generated texts.
        """
        state = self.get_cont_state(reply_to, backward)
        while True:
            parts = self.generate(state_size, state, dataset, backward)
            if reply_to is not None:
                if backward:
                    parts = chain(reversed(list(parts)), (reply_to,))
                else:
                    parts = chain((reply_to,), parts)
            parts = islice(parts, 0, max_length)
            yield self.format(parts)

    def generate_replies(self, max_length, state_size, reply_to, dataset):
        """Generate replies.

        Parameters
        ----------
        max_length : `int` or `None`
            Maximum sentence length.
        state_size : `int`
            State size.
        reply_to : `str`
            Input string.
        dataset: `str`
            Dataset key prefix.

        Returns
        -------
        `generator` of `str`
            Generated texts.
        """
        state_sets = self.get_reply_states(
            reply_to,
            dataset + state_size_dataset(state_size)
        )

        if not state_sets:
            yield from self.generate_cont(max_length, state_size,
                                          None, False, dataset)
            return

        random.shuffle(state_sets)

        generate = lambda state, backward: self.generate(
            state_size, state,
            dataset, backward
        )

        for states in cycle(state_sets):
            state = random.choice(states)
            parts = chain(
                reversed(list(generate(state, True))),
                (state,),
                generate(state, False)
            )
            parts = islice(parts, 0, max_length)
            yield self.format(parts)

    def __call__(self,
                 max_length=None,
                 state_size=None,
                 reply_to=None,
                 reply_mode=ReplyMode.END,
                 dataset=''):
        """Generate text.

        Parameters
        ----------
        max_length : `int` or `None`, optional
            Maximum sentence length (default: None).
        state_size : `int`, optional
            State size (default: parser.state_sizes[0]).
        reply_to : `str` or `None`, optional
            Input string (default: None).
        reply_mode : `markovchain.text.util.ReplyMode`, optional
            Reply mode (default: `markovchain.text.util.ReplyMode.END`)
        dataset: `str`, optional
            Dataset key prefix (default: '').

        Returns
        -------
        `str`
        """
        if reply_to is None:
            reply_mode = ReplyMode.END

        if state_size is None:
            state_size = next(iter(self.parser.state_sizes))

        if max_length is not None and max_length <= 0:
            return self.format('')

        if reply_mode == ReplyMode.REPLY:
            text = self.generate_replies(max_length, state_size,
                                         reply_to, dataset)
        else:
            backward = reply_mode == ReplyMode.START
            text = self.generate_cont(max_length, state_size,
                                      reply_to, backward, dataset)

        text = islice(text, 0, self.rank.size)
        text = self.rank(text)
        return random.choice(text)



import json
import sqlite3
from collections import deque
from itertools import chain, repeat, islice

from .base import Storage


class SqliteStorage(Storage):
    """SQLite storage.

    Attributes
    ----------
    db : `sqlite3.Connection`
        Database connection.
    cursor
        Database cursor.
    """
    def __init__(self, db=':memory:', settings=None):
        """SQLite storage constructor.

        Parameters
        ----------
        db : `str` or `sqlite3.Connection`, optional
            Database path or connection (default: ':memory:').
        settings: `dict`, optional
        """
        super().__init__(settings)
        if isinstance(db, str):
            db = sqlite3.connect(db, isolation_level='IMMEDIATE')
        self.db = db
        self.cursor = db.cursor()
        self.create_node_tables()
        self.update_main_table()
        self.cursor.execute('SELECT key, id FROM datasets')
        self.datasets = dict(self.cursor.fetchall())

    def __eq__(self, markov):
        raise NotImplementedError()
        #return super().__eq__(markov)

    def replace_state_separator(self, old_separator, new_separator):
        self.cursor.execute(
            'UPDATE nodes SET value = replace(value, ?, ?)',
            (old_separator, new_separator)
        )

    def get_dataset(self, key, create=False):
        try:
            return self.datasets[key]
        except KeyError:
            if not create:
                raise
            self.cursor.execute(
                'INSERT INTO datasets (key) VALUES (?)',
                (key,)
            )
            ret = self.cursor.lastrowid
            self.datasets[key] = ret
            return ret

    def add_links(self, links, dataset_prefix=''):
        for dataset, src, dst in links:
            src = list(src)
            source = self.get_node(self.join_state(src))
            if dst is None:
                target = None
            else:
                target = self.get_node(self.join_state(
                    chain(islice(src, 1, None), (dst,))
                ))
            dataset = self.get_dataset(dataset_prefix + dataset, True)
            self.cursor.execute(
                '''UPDATE links
                   SET count = count + 1
                   WHERE source=? AND target=? AND dataset=?''',
                (source, target, dataset)
            )
            self.cursor.execute(
                '''INSERT INTO links (dataset, source, target, value, bvalue)
                   SELECT ?, ?, ?, ?, ?
                   WHERE (SELECT Changes() = 0)''',
                (dataset, source, target, dst, src[0])
            )

    def get_state(self, state, size):
        state = deque(chain(repeat('', size), state), maxlen=size)
        self.cursor.execute(
            'SELECT id FROM nodes WHERE value=?',
            (self.join_state(state),)
        )
        state = self.cursor.fetchone()
        if state is None:
            return None
        return state[0]

    def get_states(self, dataset, string):
        dataset = self.get_dataset(dataset)
        self.cursor.execute(
            'SELECT DISTINCT nodes.value'
            ' FROM nodes'
            ' INNER JOIN links ON links.source = nodes.id AND links.dataset = ?'
            ' WHERE nodes.value LIKE ?',
            (dataset, '%%%s%%' % string)
        )
        ret = self.cursor.fetchall()
        for i, row in enumerate(ret):
            ret[i] = row[0]
        return ret

    def get_links(self, dataset, state, backward=False):
        if backward:
            query = ('SELECT count, bvalue, source'
                     ' FROM links'
                     ' WHERE dataset=? AND target=?')
        else:
            query = ('SELECT count, value, target'
                     ' FROM links'
                     ' WHERE dataset=? AND source=?')
        self.cursor.execute(query, (dataset, state))
        return self.cursor.fetchall()

    def follow_link(self, link, state, backward=False):
        return link[2]

    def get_tables(self):
        """Get all table names.

        Returns
        -------
        `set` of `str`
        """
        self.cursor.execute(
            'SELECT name FROM sqlite_master WHERE type="table"'
        )
        return set(x[0] for x in self.cursor.fetchall())

    def get_node(self, value):
        """Get node ID by value.

        If a node with the specified value does not exist,
        create it and return its ID.

        Parameters
        ----------
        value : `str`
            Node value.

        Returns
        -------
        `int`
            Node ID.
        """
        while True:
            self.cursor.execute(
                'SELECT id FROM nodes WHERE value=?',
                (value,)
            )
            node = self.cursor.fetchone()
            if node is not None:
                return node[0]
            self.cursor.execute(
                'INSERT INTO nodes (value) VALUES (?)',
                (value,)
            )

    def update_main_table(self):
        """Write generator settings to database.
        """
        data = (json.dumps(self.settings),)
        self.cursor.execute('''
            CREATE TABLE IF NOT EXISTS main (
                settings TEXT NOT NULL DEFAULT "{}"
            )
        ''')
        self.cursor.execute('SELECT * FROM main')
        if self.cursor.fetchall() == []:
            self.cursor.execute('INSERT INTO main (settings) VALUES (?)', data)
        else:
            self.cursor.execute('UPDATE main SET settings=?', data)

    def create_node_tables(self):
        """Create node and link tables if they don't exist.
        """
        self.cursor.execute('PRAGMA foreign_keys=1')
        self.cursor.execute('''
            CREATE TABLE IF NOT EXISTS datasets (
                id INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
                key TEXT NOT NULL
            )
        ''')
        self.cursor.execute('''
            CREATE TABLE IF NOT EXISTS nodes (
                id INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
                value TEXT NOT NULL
            )
        ''')
        self.cursor.execute('''
            CREATE TABLE IF NOT EXISTS links (
                dataset REFERENCES datasets (id),
                source REFERENCES nodes (id),
                target REFERENCES nodes (id),
                value TEXT,
                bvalue TEXT,
                count INTEGER NOT NULL DEFAULT 1
            )
        ''')
        self.cursor.execute(
            'CREATE UNIQUE INDEX IF NOT EXISTS node ON nodes (value)'
        )
        self.cursor.execute(
            'CREATE INDEX IF NOT EXISTS link_source ON links (source, dataset)'
        )
        self.cursor.execute(
            'CREATE INDEX IF NOT EXISTS link_target ON links (target, dataset)'
        )

    def do_save(self, fp=None):
        """Save.

        Parameters
        ----------
        fp : `None`, optional
        """
        if fp is not None:
            raise NotImplementedError()
        self.update_main_table()
        self.db.commit()

    @classmethod
    def load(cls, fp):
        if not isinstance(fp, sqlite3.Connection):
            if not isinstance(fp, str):
                fp.close()
                fp = fp.name
            fp = sqlite3.connect(fp, isolation_level='IMMEDIATE')

        cursor = fp.cursor()
        try:
            cursor.execute('SELECT settings FROM main')
            settings = json.loads(cursor.fetchone()[0])
        except sqlite3.OperationalError:
            settings = None

        return cls(fp, settings)



import os
import pytest

from markovchain import SqliteStorage


def get_nodes(cursor):
    cursor.execute('SELECT id, value FROM nodes')
    return cursor.fetchall()

def get_datasets(cursor):
    cursor.execute('SELECT id, key FROM datasets')
    return cursor.fetchall()

def get_links(cursor, source):
    cursor.execute(
        'SELECT dataset, value, count FROM links WHERE source=?',
        (source,)
    )
    return cursor.fetchall()

def test_sqlite_storage_empty():
    storage = SqliteStorage()
    assert storage.db
    assert storage.cursor
    tables = storage.get_tables()
    assert 'main' in tables
    assert 'nodes' in tables
    assert 'links' in tables
    assert 'datasets' in tables

def test_sqlite_storage_get_dataset():
    storage = SqliteStorage()
    with pytest.raises(KeyError):
        storage.get_dataset('0')
    assert storage.get_dataset('0', True) == 1
    assert storage.get_dataset('0', True) == 1
    assert storage.get_dataset('1', True) == 2

def test_sqlite_storage_add_links():
    storage = SqliteStorage()
    storage.add_links([
        ('0', ('x',), 'y'),
        ('0', ('y',), 'z'),
        ('0', ('x',), 'y')
    ])
    assert get_datasets(storage.cursor) == [(1, '0')]
    assert get_nodes(storage.cursor) == [(1, 'x'), (2, 'y'), (3, 'z')]
    assert get_links(storage.cursor, 1) == [(1, 'y', 2)]
    assert get_links(storage.cursor, 2) == [(1, 'z', 1)]
    assert get_links(storage.cursor, 3) == []

    storage.add_links([('0', ('x',), 'z'), ('1', ('x',), 'y')])
    assert get_datasets(storage.cursor) == [(1, '0'), (2, '1')]
    assert get_links(storage.cursor, 1) == [
        (1, 'y', 2), (1, 'z', 1), (2, 'y', 1)
    ]

@pytest.mark.parametrize('state,size,res', [
    ([], 4, None),
    (['x'], 1, 1),
    (['x', 'y'], 1, 2),
    (['x', 'y'], 2, 3)
])
def test_sqlite_storage_get_state(state, size, res):
    storage = SqliteStorage()
    storage.add_links([('0', ('x',), 'y'), ('0', ('x', 'y'), 'z')])
    assert storage.get_state(state, size) == res

@pytest.mark.parametrize('dataset,string,res', [
    ('0', 'x', ['XX', 'xy', 'xz']),
    ('0', 'y', ['xy', 'yz']),
    ('0', 'q', []),
    ('1', 'x', ['x']),
    ('1', 'y', [])
])
def test_json_storage_get_states(dataset, string, res):
    storage = SqliteStorage()
    storage.add_links([
        ('0', ('XX',), 'xy'),
        ('0', ('xy',), 'yz'),
        ('0', ('yz',), 'xz'),
        ('0', ('xz',), None),
        ('1', ('x',), 'y')
    ])
    test = storage.get_states(dataset, string)
    test.sort()
    assert test == res

@pytest.mark.parametrize('args,res', [
    ((1,), [(1, 'y', 2), (1, 'z', 3)]),
    ((2,), [(1, 'z', 3)]),
    ((3,), []),
    ((1, True), []),
    ((2, True), [(1, 'x', 1)]),
    ((3, True), [(1, 'x', 1), (1, 'y', 2)])
])
def test_sqlite_storage_get_links(args, res):
    storage = SqliteStorage()
    storage.add_links([
        ('0', ('x',), 'y'),
        ('0', ('x',), 'z'),
        ('0', ('y',), 'z')
    ])
    assert storage.get_links(1, *args) == res

@pytest.mark.parametrize('args,res', [
    (((0, 1, 2), 3, False), 2),
    (((1, 2, 3), 4, True), 3),
])
def test_sqlite_storage_follow_link(args, res):
    storage = SqliteStorage()
    assert storage.follow_link(*args) == res

def test_sqlite_storage_state_separator():
    storage = SqliteStorage()
    storage.add_links([('0', ('x', 'y'), 'z'), ('0', ('y', 'z'), 'x')])
    assert get_nodes(storage.cursor) == [(1, 'x y'), (2, 'y z'), (3, 'z x')]
    storage.state_separator = ':'
    assert get_nodes(storage.cursor) == [(1, 'x:y'), (2, 'y:z'), (3, 'z:x')]

def test_sqlite_storage_save_load(tmpdir):
    db = os.path.join(str(tmpdir), 'test.db')
    storage = SqliteStorage(db=db)
    storage.state_separator = ':'
    storage.add_links([('0', ('x', 'y'), 'z'), ('0', ('y', 'z'), 'x')])
    nodes = get_nodes(storage.cursor)
    storage.save()
    loaded = SqliteStorage.load(db)
    assert nodes == get_nodes(loaded.cursor)
    assert storage.state_separator == loaded.state_separator



from .markov import MarkovText
from .formatter import FormatterBase, Noop as NoopFormatter, Formatter
from .rank import Rank, Const as ConstRank, Test as TestRank
from .scanner import CharScanner, RegExpScanner
from .util import ReplyMode
from ..scanner import Scanner


FormatterBase.add_class(NoopFormatter, Formatter)
Rank.add_class(ConstRank, TestRank)
Scanner.add_class(CharScanner, RegExpScanner)



from PIL import Image

from .util import convert, palette as default_palette
from ..util import SaveLoad, DOC_INHERIT


class ImageType(SaveLoad, metaclass=DOC_INHERIT):
    """Base image type class.

    Attributes
    ----------
    classes : `dict`
        Image traversal class group.
    mode : `str`
        Image mode.
    channel_mode : `str` or None
        Channel image mode.
    channels : `list` of `str`
        Channel dataset keys.
    """

    classes = {}
    mode = None
    channel_mode = None
    channels = ['']

    def __eq__(self, imgtype):
        return self.__class__ is imgtype.__class__

    def convert(self, img):
        """Convert an image to type.

        Parameters
        ----------
        img : `PIL.Image.Image`

        Returns
        -------
        `tuple` of `PIL.Image.Image`
        """
        return (img.convert(self.mode),)

    def create(self, width, height):
        """Create an image of type.

        Parameters
        ----------
        width: `int`
            Image width.
        height: `int`
            Image height.

        Returns
        -------
        `PIL.Image.Image`
        """
        return Image.new(self.mode, (width, height))

    def create_channel(self, width, height):
        """Create a channel.

        Parameters
        ----------
        width: `int`
            Image width.
        height: `int`
            Image height.

        Returns
        -------
        `PIL.Image.Image`
        """
        if self.channel_mode is None:
            return self.create(width, height)
        return Image.new(self.channel_mode, (width, height))

    def merge(self, imgs):
        """Merge image channels.

        Parameters
        ----------
        imgs : `list` of `PIL.Image.Image`

        Returns
        -------
        `PIL.Image.Image`

        Raises
        ------
        ValueError
            If image channel list is empty.
        """
        if not imgs:
            raise ValueError('empty channel list')
        if len(imgs) == 1:
            return imgs[0]
        return Image.merge(self.mode, imgs)


class Grayscale(ImageType):
    """Grayscale image type.
    """
    mode = 'L'


class RGB(ImageType):
    """RGB image type.
    """

    mode = 'RGB'
    channel_mode = 'L'
    channels = ['_R', '_G', '_B']

    def convert(self, img):
        return img.convert(self.mode).split()


class Indexed(ImageType):
    """Indexed image type.

    Attributes
    ----------
    palette : `list` of `int`
    dither : `bool`
    convert_type : `int`
    """

    mode = 'P'
    channel_mode = None

    def __init__(self, palette=None, dither=True, convert_type=1):
        self._palette = None
        self._palette_image = None
        self.palette = palette
        self.dither = dither
        self.convert_type = convert_type

    def __eq__(self, imgtype):
        return (super().__eq__(imgtype)
                and self.palette == imgtype.palette
                and self.dither == imgtype.dither
                and self.convert_type == imgtype.convert_type)

    @property
    def palette(self):
        return self._palette

    @palette.setter
    def palette(self, palette):
        if palette is None:
            palette = [8, 4, 8]
        if len(palette) == 3:
            palette = default_palette(*palette)
        self._palette_image = Image.new('P', (1, 1))
        self._palette_image.putpalette(palette)
        self._palette = palette

    def create(self, width, height):
        ret = super().create(width, height)
        ret.putpalette(self.palette)
        return ret

    def convert(self, img):
        ret = convert(self.convert_type, img, self._palette_image, self.dither)
        return (ret,)

    def save(self):
        ret = super().save()
        ret['palette'] = self.palette
        ret['dither'] = self.dither
        ret['convert_type'] = self.convert_type
        return ret



from unittest.mock import Mock
import pytest

from markovchain.text import MarkovText, ReplyMode
from markovchain.scanner import Scanner
from markovchain.parser import Parser
from markovchain.storage import JsonStorage


def test_markov_text_data(mocker):
    mock = mocker.patch('markovchain.Markov.data', return_value=1)
    markov = MarkovText()
    assert markov.data([1, 2], True) == 1
    mock.assert_called_once_with([1, 2], True)

@pytest.mark.parametrize('test,join_with', [
    (['1', '2', '3'], ''),
    (['1', '2', '3'], ' ')
])
def test_markov_text_format(test, join_with):
    fmt = Mock(return_value=2)
    markov = MarkovText(formatter=fmt)
    markov.storage.state_separator = join_with
    assert markov.format(test) == 2
    fmt.assert_called_with(join_with.join(test))

@pytest.mark.parametrize('ss,data,args,res', [
    (1, [], (), KeyError),
    (1, 'xy', (), ['x', 'y']),
    (1, 'xy', (None, None, 'z'), ['z']),
    (1, 'xy', (None, None, 'xyx'), ['xyx', 'y']),
    (1, 'xy', (None, None, 'xyx', ReplyMode.START), ['xyx']),
    (1, 'zxy', (None, None, 'yxy', ReplyMode.START), ['z', 'x', 'yxy']),
    (2, ['u', 'x', 'xz', 'v'], (None, None, 'y z w', ReplyMode.REPLY), ['u', 'x xz', 'v']),
    (1, 'xxxxx', (2,), ['x', 'x']),
    (1, 'xxxxx', (-10,), []),
    (1, 'xxxxx', (0,), [])
])
def test_markov_text_generate(mocker, ss, data, args, res):
    fmt = mocker.patch(
        'markovchain.text.MarkovText.format',
        wraps=list
    )
    markov = MarkovText(
        parser=Parser(state_sizes=[ss]),
        scanner=Scanner(lambda x: x),
        storage=JsonStorage(backward=True)
    )
    markov.data(data)
    if isinstance(res, type):
        with pytest.raises(res):
            markov(*args)
    else:
        assert markov(*args) == res
        assert fmt.call_count == 1



from io import StringIO
from collections import deque
import pytest

from markovchain import JsonStorage


def test_json_storage_empty():
    storage = JsonStorage()
    assert storage.nodes == {}
    assert storage.settings == {}

def test_json_storage_get_dataset():
    storage = JsonStorage()
    with pytest.raises(KeyError):
        storage.get_dataset('0')
    data = storage.get_dataset('0', True)
    assert data == ({}, None)
    assert storage.get_dataset('0', True)[0] is data[0]
    assert storage.get_dataset('1', True)[0] is not data[0]

def test_json_storage_get_dataset_backward():
    storage = JsonStorage(nodes={}, backward=True)
    data = storage.get_dataset('0', True)
    assert data == ({}, {})
    assert data[0] is not data[1]

def test_json_storage_add_links():
    storage = JsonStorage()
    storage.add_links([
        ('0', ('x',), 'y'),
        ('0', ('y',), 'z'),
        ('0', ('x',), 'y')
    ])
    assert storage.nodes == {
        '0': {
            'x': [2, 'y'],
            'y': [1, 'z']
        }
    }
    storage.add_links([
        ('0', ('z',), 'x'),
        ('0', ('x',), 'z'),
        ('0', ('x',), 'y')
    ])
    assert storage.nodes == {
        '0': {
            'x': [[3, 1], ['y', 'z']],
            'y': [1, 'z'],
            'z': [1, 'x']
        }
    }
    storage.add_links([
        ('1', ('x',), 'y')
    ])
    assert storage.nodes == {
        '0': {
            'x': [[3, 1], ['y', 'z']],
            'y': [1, 'z'],
            'z': [1, 'x']
        },
        '1': {
            'x': [1, 'y']
        }
    }
    assert storage.backward is None

def test_json_storage_add_links_backward():
    storage = JsonStorage(backward=True)
    storage.add_links([
        ('0', ('x', 'y'), 'y'),
        ('0', ('y', 'z'), 'z'),
        ('0', ('z', 'x'), 'x')
    ])
    assert storage.backward == {
        '0': {
            'y y': [1, 'x'],
            'z z': [1, 'y'],
            'x x': [1, 'z']
        }
    }

@pytest.mark.parametrize('state,size,res', [
    ([], 4, ['', '', '', '']),
    (['ab', 'cd'], 1, ['cd']),
    (['ab', 'cd'], 3, ['', 'ab', 'cd'])
])
def test_json_storage_get_state(state, size, res):
    storage = JsonStorage()
    assert storage.get_state(state, size) == deque(res, maxlen=size)

@pytest.mark.parametrize('dataset,string,res', [
    ('0', 'x', ['XX', 'xy', 'xz']),
    ('0', 'y', ['xy', 'yz']),
    ('0', 'q', []),
    ('1', 'x', ['x']),
    ('1', 'y', [])
])
def test_json_storage_get_states(dataset, string, res):
    storage = JsonStorage()
    storage.add_links([
        ('0', ('XX',), 'xy'),
        ('0', ('xy',), 'yz'),
        ('0', ('yz',), 'xz'),
        ('0', ('xz',), None),
        ('1', ('x',), 'y')
    ])
    test = storage.get_states(dataset, string)
    test.sort()
    assert test == res

@pytest.mark.parametrize('args,res', [
    ((['x'],), [(1, 'y')]),
    ((['x', 'y'],), [(1, 'z')]),
    ((['y'],), [(1, 'x'), (2, 'y'), (3, 'z')]),
    ((['z'],), []),
    ((['x'], True), [(1, 'z')]),
    ((['y'], True), [])
])
def test_json_storage_get_links(args, res):
    dataset = (
        {
            'x': [1, 'y'],
            'y': [[1, 2, 3], ['x', 'y', 'z']],
            'x y': [1, 'z']
        },
        {
            'x': [1, 'z']
        }
    )
    storage = JsonStorage(backward=True)
    assert storage.get_links(dataset, *args) == res

@pytest.mark.parametrize('args,res', [
    (((0, 'y'), deque(['x'])), ['x', 'y']),
    (((0, 'y'), deque(['x']), True), ['y', 'x']),
])
def test_json_storage_follow_link(args, res):
    storage = JsonStorage()
    assert storage.follow_link(*args) == deque(res)

def test_json_storage_state_separator():
    storage = JsonStorage(backward=True)
    storage.add_links([('0', ('x', 'y'), 'z'), ('1', ('y', 'z'), 'x')])
    assert storage.nodes == {
        '0': {
            'x y': [1, 'z']
        },
        '1': {
            'y z': [1, 'x']
        }
    }
    assert storage.backward == {
        '0': {
            'y z': [1, 'x']
        },
        '1': {
            'z x': [1, 'y']
        }
    }
    storage.state_separator = ':'
    assert storage.nodes == {
        '0': {
            'x:y': [1, 'z'],
        },
        '1': {
            'y:z': [1, 'x']
        }
    }
    assert storage.backward == {
        '0': {
            'y:z': [1, 'x']
        },
        '1': {
            'z:x': [1, 'y']
        }
    }

@pytest.mark.parametrize('test,test2,res', [
    ((), (), True),
    ((), ({}, False), True),
    ((), ({}, True), False),
    (({}, {}), ({}, True), True),
    ((), ({'0': {'x':[1, 'y']}}), False),
    (({'0': {'x':[1, 'y']}}), ({'0': {'x':[1, 'y']}}), True),
    ((), ({}, {'state_separator': ':'}), False)
])
def test_json_storage_eq(test, test2, res):
    assert (JsonStorage(*test) == JsonStorage(*test2)) == res

def test_json_storage_save_load():
    storage = JsonStorage(backward=True, settings={'state_separator': ':'})
    storage.add_links([
        ('0', ('x',), 'y'),
        ('0', ('y',), 'z'),
        ('1', ('x',), 'y'),
        ('1', ('x',), 'z')
    ])
    fp = StringIO()
    storage.save(fp)
    fp.seek(0)
    loaded = JsonStorage.load(fp)
    assert storage == loaded



from unittest.mock import Mock, call
import pytest

from markovchain import Parser, LevelParser, Scanner
from markovchain.parser import ParserBase


def parse(parser, data, part=False, separator=' '):
    return [(dataset, separator.join(src), dst)
            for dataset, src, dst in parser(data, part)]


def test_parser_properties():
    parser = Parser()

    with pytest.raises(ValueError):
        parser.state_sizes = [1, 0, 2]
    with pytest.raises(ValueError):
        parser.state_sizes = []

    assert parser.state_sizes == [1]
    assert parser.state_size == 1

    parser.state_sizes = [2, 1]
    assert parser.state_sizes == [2, 1]
    assert parser.state_size == 2
    assert parser.state.maxlen == 2
    assert list(parser.state) == ['', '']

    parser.state.append('test')
    parser.state_size = parser.state_size
    assert list(parser.state) == ['', 'test']

    parser.state_sizes = [3]
    assert list(parser.state) == ['', '', '']

def test_parser_parse(mocker):
    state_size_dataset = mocker.patch(
        'markovchain.parser.state_size_dataset',
        return_value='0'
    )
    parser = Parser(state_sizes=[2], reset_on_sentence_end=False)
    res = parse(parser, ['a', Scanner.END, 'b'], True, '::')
    assert res == [('0', '::', 'a'), ('0', '::a', None), ('0', '::a', 'b')]
    res = parse(parser, 'c', separator='::')
    assert res == [('0', 'a::b', 'c')]
    state_size_dataset.assert_has_calls(
        call(parser.state_sizes[0]) for _ in range(2)
    )

def test_parser_parse_default(mocker):
    state_size_dataset = mocker.patch(
        'markovchain.parser.state_size_dataset',
        return_value='0'
    )
    parser = Parser()
    assert parse(parser, '') == []
    assert parse(parser, 'abc', True) == [
        ('0', '', 'a'), ('0', 'a', 'b'), ('0', 'b', 'c')
    ]
    assert parse(parser, ['a', 'b', Scanner.END, 'c']) == [
        ('0', 'c', 'a'), ('0', 'a', 'b'), ('0', 'b', None), ('0', '', 'c')
    ]
    assert parse(parser, ['a', Scanner.END, Scanner.END, 'c']) == [
        ('0', '', 'a'), ('0', 'a', None), ('0', '', 'c')
    ]
    assert parse(parser, [Scanner.END] * 4) == []
    state_size_dataset.assert_has_calls(
        call(parser.state_sizes[0]) for _ in range(4)
    )

@pytest.mark.parametrize('test,res', [
    ('abcde', [('0', '  ', 'a'),
               ('0', '  a', 'b'), ('0', ' a b', 'c'),
               ('0', 'a b c', 'd'), ('0', 'b c d', 'e')]),
    (['a', 'b', 'c', Scanner.END, 'd', 'e'],
     [('0', '  ', 'a'), ('0', '  a', 'b'),
      ('0', ' a b', 'c'), ('0', 'a b c', None),
      ('0', '  ', 'd'), ('0', '  d', 'e')]),
    (['a', 'b', 'c', (Scanner.START, 'd'), 'e'],
     [('0', '  ', 'a'), ('0', '  a', 'b'),
      ('0', ' a b', 'c'), ('0', '  d', 'e')])
])
def test_parser_state_size(mocker, test, res):
    state_size_dataset = mocker.patch(
        'markovchain.parser.state_size_dataset',
        return_value='0'
    )
    parser = Parser(state_sizes=[3])
    assert parse(parser, test) == res
    state_size_dataset.assert_called_once_with(3)

@pytest.mark.parametrize('test,test2,res', [
    ((), (), True),
    (([2], False), ([2], False), True),
    (([2], False), ([2], True), False),
    (([2], False), ([2, 3], False), False)
])
def test_parser_eq(test, test2, res):
    parser = Parser(*test)
    parser2 = Parser(*test2)
    assert (parser == parser2) == res

@pytest.mark.parametrize('test', [
    ([1, 2, 3], False)
])
def test_parser_save_load(test):
    parser = Parser(*test)
    saved = parser.save()
    loaded = Parser.load(saved)
    assert parser == loaded


def test_level_parser_properties():
    parser = LevelParser()
    assert parser.levels == 1
    assert parser.parsers == [Parser()]

    with pytest.raises(ValueError):
        parser.levels = 0
    with pytest.raises(ValueError):
        parser.levels = -1
    assert parser.levels == 1

    parser.levels = 2
    assert parser.parsers == [Parser(), Parser()]

    level = Parser(state_sizes=[2, 3])
    parser.parsers = level
    assert parser.parsers == [level, level]

    parser.parsers = [Parser(), level, Parser()]
    assert parser.parsers == [Parser(), level]

    parser.levels = 1
    assert parser.parsers == [Parser()]

    parser.levels = 2
    assert parser.parsers[1] is level

def test_level_parser_reset():
    parsers = [Mock(), Mock()]
    parser = LevelParser(levels=2, parsers=parsers)
    parser.reset()
    for level in parsers:
        level.reset.assert_called_once_with()

@pytest.mark.parametrize('test,res', [
    ([[0], [1]], [0, 1]),
    ([[0]] * 5, [0, 1]),
    ([], [])
])
def test_level_parser_parse(mocker, test, res):
    level_dataset = mocker.patch(
        'markovchain.parser.level_dataset',
        return_value='0'
    )
    parsers = [
        Mock(wraps=ParserBase(lambda x: [0])),
        Mock(wraps=ParserBase(lambda x: [1]))
    ]
    parser = LevelParser(
        levels=2,
        parsers=parsers
    )
    assert list(parser(test, dataset='data')) == res
    if test:
        assert level_dataset.call_count == 2
        level_dataset.assert_has_calls([call(0), call(1)])
    for parser, data in zip(parsers, test):
        parser.assert_called_once_with(data, False, 'data0')

@pytest.mark.parametrize('test,test2,res', [
    ((), (), True),
    ((2, Parser()), (2, Parser()), True),
    ((1,), (2,), False),
    ((2, [Parser(), Parser(3)]), (2, Parser()), False)
])
def test_level_parser_eq(test, test2, res):
    assert (LevelParser(*test) == LevelParser(*test2)) == res

def test_level_parser_save_load():
    level = Parser(state_sizes=[2, 3])
    parser = LevelParser(levels=3, parsers=[level, Parser()])
    saved = parser.save()
    loaded = Parser.load(saved)
    assert parser == loaded



import pytest
from enum import IntEnum

from markovchain.util import (
    SaveLoad, ObjectWrapper, const,
    fill, load, extend, to_list, truncate,
    state_size_dataset, level_dataset, int_enum
)


@pytest.fixture
def save_load_test():
    class SaveLoadTest(SaveLoad):
        classes = {}

        def __init__(self, value):
            self.value = value

        def __eq__(self, test):
            return self.value == test.value

        def save(self):
            data = super().save()
            data['value'] = self.value
            return data
    SaveLoadTest.add_class(SaveLoadTest)
    return SaveLoadTest


def test_saveload_add_remove_class(save_load_test):
    assert save_load_test.classes['SaveLoadTest'] is save_load_test
    save_load_test.remove_class(save_load_test)
    save_load_test.remove_class(save_load_test)
    with pytest.raises(KeyError):
        save_load_test.classes['SaveLoadTest']

def test_saveload_save_load(save_load_test):
    test = save_load_test(0)
    saved = test.save()
    loaded = save_load_test.load(saved)
    assert isinstance(loaded, save_load_test)
    assert loaded == test


class ObjectTest:
    def __init__(self, x, y):
        self.x = x
        self.y = y
    def method(self):
        return self.x + self.y
    def method2(self):
        return self.x - self.y

class ObjectWrapperTest(ObjectWrapper):
    def __init__(self, obj, z):
        super().__init__(obj)
        self.y *= 2
        self.z = z
    def method(self):
        return super().method() + self.z

def test_object_wrapper_wrap():
    obj = ObjectTest(1, 2)
    wrapped = ObjectWrapper(obj)
    assert isinstance(wrapped, ObjectTest)
    assert wrapped.__dict__ == obj.__dict__
    assert wrapped.method() == 3
    assert wrapped.method2() == -1

def test_object_wrapper_override():
    obj = ObjectTest(1, 2)
    wrapped = ObjectWrapperTest(obj, 3)
    assert wrapped.x == 1
    assert wrapped.y == 4
    assert wrapped.z == 3
    assert wrapped.method() == 8
    assert wrapped.method2() == -3


class IntEnumTest(IntEnum):
    X = 0
    Y = 1

@pytest.mark.parametrize('test,res', [
    (0, 0),
    ('Y', 1),
    ('y', 1),
    ('z', ValueError),
    (2, ValueError)
])
def test_int_enum(test, res):
    if isinstance(res, type):
        with pytest.raises(res):
            int_enum(IntEnumTest, test)
    else:
        test = int_enum(IntEnumTest, test)
        assert isinstance(test, IntEnumTest)
        assert test == res


def test_const():
    assert const(0)() == 0
    assert const(1)(1, [2], key=3) == 1


@pytest.mark.parametrize('test,res', [
    ([], []),
    (range(3), list(range(3))),
    (0, [0]),
    ({'x': 0}, [{'x': 0}])
])
def test_to_list(test, res):
    assert to_list(test) == res


def test_fill_error():
    with pytest.raises(ValueError):
        fill([], 1)

@pytest.mark.parametrize('test,res', [
    ((None, -1), []),
    (([], 0), []),
    ((0, 0), []),
    ((1, 1), [1]),
    ((1, 5), [1] * 5),
    ((range(10), 2), [0, 1]),
    ((range(3), 5), [0, 1, 2, 2, 2])
])
def test_fill(test, res):
    assert fill(*test) == res

@pytest.mark.parametrize('lst,size', [
    ([], 0),
    (list(range(3)), 3)
])
def test_fill_no_copy(lst, size):
    assert fill(lst, size) is lst

@pytest.mark.parametrize('lst,size', [
    ([[0], [1]], 4)
])
def test_fill_copy(lst, size):
    res = fill(lst, size, copy=False)
    assert res[-1] is lst[-1]
    res = fill(lst, size, copy=True)
    assert res[-1] is not lst[-1]
    assert res[-1] == lst[-1]


class LoadTest:
    @staticmethod
    def load(data):
        return data['x']

def test_load_efault():
    x = load(None, LoadTest, lambda: 0)
    assert x == 0

def test_load_class():
    x = load({'x': 1}, LoadTest, lambda: 0)
    assert x == 1

def test_load_object():
    obj = object()
    x = load(obj, LoadTest, lambda: 0)
    assert x is obj


@pytest.mark.parametrize('test,res', [
    (({'x': 0}, {'y': 1}), {'x': 0, 'y': 1}),
    (({'x': 0}, {'y': 1}, {'x': 1}), {'x': 1, 'y': 1}),
    (({'x': {'y': 0}}, {'x': {'z': 1}}), {'x': {'y': 0, 'z': 1}}),
    (({'x': {'y': 0}}, {'x': 1}), {'x': 1}),
    (({'x': 1}, {'x': {'y': 0}}), {'x': {'y': 0}}),
    (({}, {'x': {'y': 0}}), {'x': {'y': 0}})
])
def test_extend(test, res):
    assert extend(*test) == res


@pytest.mark.parametrize('test', [
    ('', 3, True),
    ('0', -1, False)
])
def test_truncate_error(test):
    with pytest.raises(ValueError):
        truncate(*test)

@pytest.mark.parametrize('test', [
    ('0', 4, True),
    ('1234', 4, False)
])
def test_truncate_noop(test):
    assert truncate(*test) is test[0]

@pytest.mark.parametrize('test,res', [
    (('1234567', 5), '12...'),
    (('1234567', 5, True), '12...'),
    (('1234567', 5, False), '...67')
])
def test_truncate(test, res):
    assert truncate(*test) == res


@pytest.mark.parametrize('test', range(5))
def test_state_size_dataset(test):
    res = state_size_dataset(test)
    assert isinstance(res, str)
    assert len(res) > 0
    assert res != state_size_dataset(test - 1)

@pytest.mark.parametrize('test', range(5))
def test_level_dataset(test):
    res = level_dataset(test)
    assert isinstance(res, str)
    assert len(res) > 0
    assert res != level_dataset(test - 1)
    assert res != state_size_dataset(test)



from .markov import MarkovImage
from .scanner import ImageScanner
from .traversal import Traversal, HLines, VLines, Spiral, Blocks, Hilbert
from .type import ImageType, RGB, Grayscale, Indexed

from ..scanner import Scanner
from ..parser import Parser, LevelParser

Scanner.add_class(ImageScanner)
Traversal.add_class(HLines, VLines, Spiral, Blocks, Hilbert)
ImageType.add_class(RGB, Grayscale, Indexed)



import pytest

from markovchain.text.formatter import Noop, Formatter


@pytest.mark.parametrize('test', [
    'woRd WORD word'
])
def test_noop_formatter(test):
    assert Noop()(test) is test


@pytest.mark.parametrize('kwargs,kwargs2,res', [
    ({}, {}, True),
    ({'default_end': '/'}, {}, False),
    ({'default_end': '/'}, {'default_end': '/'}, True),
    ({'end_chars': 'd'}, {}, False),
    ({'end_chars': 'd'}, {'end_chars': 'd'}, True),
    ({'case': 'upper'}, {}, False),
    ({'case': 'upper'}, {'case': 'upper'}, True),
    ({'replace': [('.', '')]}, {}, False),
    ({'replace': [('.', '')]}, {'replace': [('.', '')]}, True)
])
def test_formatter_eq(kwargs, kwargs2, res):
    assert (Formatter(**kwargs) == Formatter(**kwargs2)) == res


@pytest.mark.parametrize('kwargs', [
    {},
    {'default_end': '/'},
    {'end_chars': 'd'},
    {'case': 'upper'},
    {'replace': [('.', '')]}
])
def test_formatter_save_load(kwargs):
    fmt = Formatter(**kwargs)
    saved = fmt.save()
    loaded = Formatter.load(saved)
    assert fmt == loaded


@pytest.mark.parametrize('kwargs,test,res', [
    ({}, '', ''),
    ({}, '  ', ''),
    ({}, '  ...', ''),
    ({}, '.?!word', 'Word.'),
    ({}, 'word  ,  (word)..  word', 'Word, (word).. word.'),
    ({}, 'word,wo[rd..wo]rd', 'Word, wo [rd.. wo] rd.'),
    ({}, 'wo--*--rd', 'Wo --*-- rd.'),
    ({'default_end': '/'}, 'word', 'Word/'),
    ({'default_end': None}, 'word', 'Word'),
    ({'end_chars': 'd'}, 'word', 'Word'),
    ({'case': 'upper'}, 'word', 'WORD.'),
    ({'case': 'lower'}, 'woRd', 'word.'),
    ({'case': 'preserve'}, 'woRd', 'woRd.'),
    ({'case': 'xxx'}, 'woRd', ValueError),
    ({'replace': [(r'(x+)', r'_\1_')]}, 'axxb', 'A_xx_b.')
])
def test_formatter_format(kwargs, test, res):
    if isinstance(res, str):
        assert Formatter(**kwargs)(test) == res
    else:
        with pytest.raises(res):
            Formatter(**kwargs)(test)



import re
from abc import abstractmethod

from ..util import SaveLoad, int_enum
from .util import (
    CharCase, lstrip_ws_and_chars,
    re_flags, re_flags_str, re_sub
)


class FormatterBase(SaveLoad):
    """Text formatter base class."""
    classes = {}

    @abstractmethod
    def __call__(self, string):
        """Format a string.

        Parameters
        ----------
        string : `str`
            String to format.

        Returns
        -------
        `str`
            Formatted string.
        """
        pass


class Noop(FormatterBase):
    """No-op formatter."""
    def __call__(self, string):
        return string


class Formatter(FormatterBase):
    """Default formatter.

    Attributes
    ----------
    case : `markovchain.text.util.CharCase`
        Character case.
    replace : `list` of (_sre.SRE_Pattern, `str`, `int`)
        List of regular expressions to replace.
    end_chars : `str`
        Sentence ending characters.
    default_end : `None` or `str`
        Default sentence ending character.
    """
    # pylint:disable=bad-whitespace
    DEFAULT_REPLACE = [
        (r'\s+'               , r' '       ),
        (r'\s*([^\w\s]+)\s*'  , r'\1'      ),
        (r'([,.?!])(\w)'      , r'\1 \2'   ),
        (r'([\w,.?!])([[({<])', r'\1 \2'   ),
        (r'([])}>])(\w)'      , r'\1 \2'   ),
        (r'(\w)([-+*]+)(\w)'  , r'\1 \2 \3'),
    ]
    # pylint:enable=bad-whitespace
    def __init__(self,
                 case=CharCase.TITLE,
                 replace=None,
                 end_chars='.?!',
                 default_end='.'):
        """Formatter constructor.

        Parameters
        ----------
        case : `int` or `str` or `markovchain.text.util.CharCase`, optional
            Character case (default: `markovchain.text.util.CharCase.TITLE`).
        end_chars : `str`, optional
            Sentence ending characters (default: '.?!').
        default_end : `None` or `str`, optional
            Default sentence ending character (default: '.').
        replace : `list` of ((`str`, `str`) or (`str`, `str`, `str`)), optional
            List of regular expressions to replace (default: DEFAULT_REPLACE).
        """
        if replace is None:
            replace = self.DEFAULT_REPLACE
        self.case = int_enum(CharCase, case)
        self.end_chars = end_chars
        self.default_end = default_end
        self.replace = []
        for rule in replace:
            try:
                expr, repl, flags = rule
            except ValueError:
                expr, repl = rule
                flags = 'u'
            flags, custom_flags = re_flags(flags)
            self.replace.append((re.compile(expr, flags), repl, custom_flags))

    def save(self):
        data = super().save()
        data['case'] = self.case.name
        data['replace'] = [
            (expr.pattern, repl, re_flags_str(expr.flags, flags))
            for expr, repl, flags in self.replace
        ]
        data['end_chars'] = self.end_chars
        data['default_end'] = self.default_end
        return data

    def __eq__(self, fmt):
        return (
            self.__class__ is fmt.__class__
            and self.case == fmt.case
            and self.replace == fmt.replace
            and self.end_chars == fmt.end_chars
            and self.default_end == fmt.default_end
        )

    def __call__(self, string):
        string = lstrip_ws_and_chars(string.rstrip(), self.end_chars)

        if not string:
            return string

        if self.default_end is not None and string[-1] not in self.end_chars:
            string += self.default_end

        string = self.case.convert(string)

        for expr, repl, flags in self.replace:
            string = re_sub(expr, repl, string, custom_flags=flags)

        return string



from unittest.mock import Mock, MagicMock, mock_open

from io import StringIO
from argparse import Namespace
import json
import sys

import pytest

from markovchain import JsonStorage, SqliteStorage
from markovchain.cli.util import (
    no_tqdm, NoProgressBar,
    pprint, load, save,
    set_args, JSON, SQLITE,
    check_output_format,
    infiles, outfiles
)


def test_no_progress_bar_warning(mocker):
    stderr = mocker.patch('sys.stderr', new_callable=StringIO)
    pbar = NoProgressBar()
    assert stderr.getvalue() == ''
    assert not NoProgressBar.warning
    pbar.print_warning()
    assert stderr.getvalue() != ''
    assert NoProgressBar.warning
    pbar.print_warning()
    assert stderr.getvalue() != ''
    assert NoProgressBar.warning

def test_no_tqdm(mocker):
    no_pbar = mocker.patch('markovchain.cli.util.NoProgressBar')
    pbar = no_tqdm(total=100, leave=False, desc='')
    assert isinstance(pbar, Mock)
    no_pbar.assert_called_with()
    no_pbar.print_warning.assert_called_with()
    no_pbar.reset_mock()
    iterable = range(3)
    pbar = no_tqdm(iterable)
    assert pbar is iterable
    no_pbar.assert_not_called()
    no_pbar.print_warning.assert_called_with()


@pytest.mark.parametrize('test', [
    {'x': 0, 'y': [{'z': '0'}, {'z': '1'}]},
    [{'x': 0.1, 'y': {'y': False, 'z': None, 'u': 3}}, {'z': 0}]
])
def test_pprint(mocker, test):
    stdout = mocker.patch('sys.stdout', new_callable=StringIO)
    res = json.dumps(test, indent=4, sort_keys=True) + '\n'
    pprint(test)
    assert stdout.getvalue() == res

@pytest.mark.parametrize('test,res', [
    ([], '[]\n'),
    ([0, 1, 2], '[0, 1, 2]\n'),
    ([True, None], '[true, null]\n'),
    ([0, [1, 2], 3], '[\n    0,\n    [1, 2],\n    3\n]\n')
])
def test_pprint_array(mocker, test, res):
    stdout = mocker.patch('sys.stdout', new_callable=StringIO)
    pprint(test)
    assert stdout.getvalue() == res


@pytest.mark.parametrize('fname,bz2,stdout', [
    ('test.json', False, False),
    ('test.json.bz2', True, True)
])
def test_load_json(mocker, fname, bz2, stdout):
    open_ = mocker.patch('builtins.open', new_callable=mock_open)
    bz2open = mocker.patch('bz2.open', new_callable=mock_open)
    stdout_ = mocker.patch('sys.stdout', new_callable=StringIO)
    json_storage = MagicMock()
    json_storage_cls = mocker.patch(
        'markovchain.cli.util.JsonStorage',
        load=Mock(return_value=json_storage)
    )
    if bz2:
        handle = bz2open()
    else:
        handle = open_()

    cls = Mock(from_storage=Mock(return_value=0))
    args = Namespace(type=JSON, progress=stdout, settings={})

    assert load(cls, fname, args) == 0
    assert (stdout_.getvalue() != '') == stdout
    if bz2:
        assert not open_.called
        bz2open.assert_called_with(fname, 'rt')
    else:
        assert not bz2open.called
        open_.assert_called_with(fname, 'rt')
    json_storage_cls.load.assert_called_once_with(handle)
    cls.from_storage.assert_called_once_with(json_storage)

def test_load_sqlite(mocker):
    sqlite_storage = MagicMock()
    sqlite_storage_cls = mocker.patch(
        'markovchain.cli.util.SqliteStorage',
        load=Mock(return_value=sqlite_storage)
    )
    fname = 'test'
    cls = Mock(from_storage=Mock(return_value=0))
    args = Namespace(type=SQLITE, progress=False, settings={})
    assert load(cls, fname, args) == 0
    sqlite_storage_cls.load.assert_called_once_with(fname)
    cls.from_storage.assert_called_once_with(sqlite_storage)


@pytest.mark.parametrize('fname,bz2,stdout', [
    ('test.json', False, False),
    ('test.json.bz2', True, True)
])
def test_save_json(mocker, fname, bz2, stdout):
    open_ = mocker.patch('builtins.open', new_callable=mock_open)
    bz2open = mocker.patch('bz2.open', new_callable=mock_open)
    stdout_ = mocker.patch('sys.stdout', new_callable=StringIO)
    markov = Mock(storage=JsonStorage(), save=Mock())

    if bz2:
        handle = bz2open()
    else:
        handle = open_()

    args = Namespace(progress=stdout, settings={})

    save(markov, fname, args)
    assert (stdout_.getvalue() != '') == stdout
    if bz2:
        assert not open_.called
        bz2open.assert_called_with(fname, 'wt')
    else:
        assert not bz2open.called
        open_.assert_called_with(fname, 'wt')
    markov.save.assert_called_once_with(handle)

def test_save_sqlite():
    markov = Mock(storage=SqliteStorage(), save=Mock())
    fname = 'test'
    args = Namespace(progress=False, settings={})
    save(markov, fname, args)
    markov.save.assert_called_once_with()


def test_set_args_error():
    args = Namespace(output=sys.stdout, progress=True)
    with pytest.raises(ValueError):
        set_args(args)

@pytest.mark.parametrize('args,res', [
    (Namespace(), JSON),
    (Namespace(output='.db'), SQLITE),
    (Namespace(state='.json'), JSON),
    (Namespace(state='.json.bz2'), JSON),
    (Namespace(type='json'), JSON),
    (Namespace(output='.db', state='.json.bz2'), JSON),
    (Namespace(output='.db', state='.json.bz2', type='sqlite'), SQLITE)
])
def test_set_args_type(args, res):
    set_args(args)
    assert args.type == res

def test_set_args_settings(mocker):
    mocker.patch('json.load', new=lambda x: x())

    args = Namespace()
    set_args(args)
    assert args.settings == {} # pylint:disable=no-member

    args = Namespace(settings=None)
    set_args(args)
    assert args.settings == {} # pylint:disable=no-member

    s = list(range(3))
    mock = MagicMock(return_value=s)
    args = Namespace(settings=mock)
    set_args(args)
    assert args.settings == s # pylint:disable=no-member
    mock.close.assert_called_with()


@pytest.mark.parametrize('test', [
    ('test', -1),
    ('test', 0),
    ('test', 2),
    ('test%d%d', 2)
])
def test_check_output_format_error(test):
    with pytest.raises(ValueError):
        check_output_format(*test)

@pytest.mark.parametrize('test', [
    ('test', 1),
    ('test%d', 2)
])
def test_check_output_format(test):
    check_output_format(*test)


@pytest.mark.parametrize('test', [
    ([], True),
    (list(range(3)), False)
])
def test_infiles(mocker, test):
    tqdm = mocker.patch('markovchain.cli.util.tqdm')
    with infiles(*test) as iter_:
        assert not tqdm.called
        assert iter_ is test[0]

def test_infiles_progress(mocker):
    tqdm = mocker.patch('markovchain.cli.util.tqdm')
    test = [0, 1]
    with infiles(test, True) as pbar:
        assert tqdm.called
        assert (test,) in tqdm.call_args
        assert isinstance(pbar, Mock)
    pbar.close.assert_called_once_with()


@pytest.mark.parametrize('test', [
    ('', -1, False),
    ('', 0, False)
])
def test_outfiles_error(test):
    with pytest.raises(ValueError), outfiles(*test):
        pass

@pytest.mark.parametrize('test,res', [
    (('', 1, False), ['']),
    (('%d', 2, False), ['0', '1'])
])
def test_outfiles(mocker, test, res):
    tqdm = mocker.patch('markovchain.cli.util.tqdm')
    with outfiles(*test) as files:
        files = list(files)
        assert not tqdm.called
        assert files == res

def test_outfiles_progress(mocker):
    tqdm = mocker.patch('markovchain.cli.util.tqdm')
    test = ('%d', 3, True)
    res = ['0', '1', '2']
    with outfiles(*test) as pbar:
        assert tqdm.called
        assert list(tqdm.call_args[0][0]) == res
        assert isinstance(pbar, Mock)
    pbar.close.assert_called_once_with()



import os
import json
import pytest

from markovchain.cli.main import main


@pytest.mark.parametrize('fname,settings,data,args,res', [
])
def test_cli_image(mocker, mock_cli, fname, settings, data, args, res):
    mock_cli(mocker)

    statefile = os.path.join(mock_cli.dir, fname)
    datafile = os.path.join(mock_cli.dir, 'data.txt')
    settingsfile = os.path.join(mock_cli.dir, 'settings.json')

    cmd = ['text', 'create', '-o', statefile]
    if settings is not None:
        with open(settingsfile, 'wt') as fp:
            json.dump(settings, fp)
        cmd.extend(('-s', settingsfile))
    if len(data) > 0:
        with open(datafile, 'wt') as fp:
            fp.write(data[0])
        cmd.append(datafile)
    mock_cli.run(main, cmd)
    mock_cli.assert_output('', '')

    update = data[1:]
    if update:
        cmd = ['text', 'update']
        cmd.append(statefile)
        for i, data_ in enumerate(data):
            datafile = os.path.join(mock_cli.dir, 'data%d.txt' % i)
            cmd.append(datafile)
            with open(datafile, 'wt') as fp:
                fp.write(data_)
        mock_cli.run(main, cmd)
        mock_cli.assert_output('', '')

    cmd = ['text', 'generate']
    cmd.extend(args)
    cmd.append(statefile)
    mock_cli.run(main, cmd)
    mock_cli.assert_output(res, '')
